-- POWERS 3
-- irons

Triggers = {}

CollectionsUsed = {16}

local magicalDamageType = "staff"
local magicalDamageAmount = 1
local anchorEffect = "minor fusion detonation"
local playerPushPull, playerRiseFall = 0.007, 0.009
local playerTangentSpeed = 0.003
local anchorTracer = "fusion bolt minor"
local connectionSegments = 10
local connectionEffect = "compiler bolt major contrail"

local iconCountdown = 10 -- ticks

local justPulled = 0
local justChoseSurface = 1
local justMelded = 2
local justMovedCloser = 3
local justMovedFarther = 4
local justReeled = 5

local categories = {}

categories[justPulled] = justPulled
categories[justChoseSurface] = justPulled

categories[justMelded] = justMelded
categories[justMovedCloser] = justMelded
categories[justMovedFarther] = justMelded
categories[justReeled] = justMelded

function Triggers.init()
   for p in Players() do
      p._currentActions = {}
   end
end

function Triggers.idle()
   for p in Players() do
      p._pulling = false
      targets(p)
      powers(p)
      updateMonsterAnchor(p)
      showHUD(p)
   end
end

function Triggers.monster_damaged(victim, aggressor, damageType,
				  damageAmount, projectile)
   if not victim.player and not aggressor
      and damageType == magicalDamageType and
      damageAmount == magicalDamageAmount then
      victim.life = victim.life + magicalDamageAmount
   end
end

function Triggers.monster_killed(victim, aggressor, projectile)
   unanchorMonster(victim)
   if aggressor and victim and aggressor._melded == victim then
      removeMelding(aggressor)
   end
end

function powers(player)
   if not player.dead then
      push(player)
      pull(player)
      adjustMeldingDistance(player)
      player.weapons.active = false
   end
end

function adjustMeldingDistance(player)
   local secondary = player.action_flags.right_trigger
   
   if player._melding and player._melding.valid then
      local direction = false

      if player.action_flags.cycle_weapons_forward then
	 player.action_flags.cycle_weapons_forward = false
	 player._meldingAdjustmentCoefficient = 1
	 player._currentActions[justMovedFarther] = iconCountdown
	 direction = true
      end
      
      if player.action_flags.cycle_weapons_backward then
	 player.action_flags.cycle_weapons_backward = false
	 player._meldingAdjustmentCoefficient = -1
	 player._currentActions[justMovedCloser] = iconCountdown
	 direction = true
      end

      local coeff = player._meldingAdjustmentCoefficient
      
      if coeff then
	 
	 local dist
	 
	 if secondary then
	    dist = coeff * 0.25
	    player._meldingDistance = player._meldingDistance + dist
	    player._currentActions[justReeled] = iconCountdown
	 elseif direction then
	    dist = coeff * 0.5
	    player._meldingDistance = player._meldingDistance + dist
	 end

	 if dist and player._meldingDistance <= dist then
	    player._meldingDistance = dist
	 end
	 
      end
   end
end

function push(player)
   if not player._melding then
      if player.action_flags.right_trigger then
	 player.action_flags.right_trigger = false
	 if player._anchor then
	    anchorLogic(player)
	 end
      end
   end
end

function pull(player)
   if player.action_flags.left_trigger then
      player.action_flags.left_trigger = false
      if player._anchor then
	 pullTowardsAnchor(player)
	 illustrateConnection(player)
      else
	 motion(player)
      end
   else
      if not player._anchor then removeMelding(player) end
   end
end

function pullTowardsAnchor(player)
   local yaw, pitch = findAnglesToAnchor(player)
   while (yaw < 0) do yaw = yaw + 360 end
   player:accelerate(yaw, playerPushPull, pitch/90 * playerRiseFall)
   player._pulling = true
   player._currentActions[justPulled] = iconCountdown
end

function findAnglesToAnchor(player)
   return normalizeAngle(findYawToAnchor(player)), findPitchToAnchor(player)

end

function findYawToAnchor(player)
   return math.deg(math.atan2(player._anchor.y - player.monster.y,
			      player._anchor.x - player.monster.x))
end

function findPitchToAnchor(player)
   local opposite = player._anchor.z - (player.monster.z + 0.6)
   local adjacent =
      distance2D(player._anchor.x, player._anchor.y,
		 player.monster.x, player.monster.y)
   return math.deg(math.atan2(opposite, adjacent))
end

function distance2D(x1, y1, x2, y2)
   return math.sqrt((x1 - x2)^2 + (y1 - y2)^2)
end

function distance3D(x1, y1, z1, x2, y2, z2)
   return math.sqrt((x1 - x2)^2 + (y1 - y2)^2 + (z1 - z2)^2)
end

function targets(player)
   if player.action_flags.action_trigger and
      not player:find_action_key_target() and not player.dead then
      if not player._anchor then
	 anchorLogic(player)
      else
	 resetAnchor(player)
      end
   end
end

function anchorLogic(player)
   local target, x, y, z, polygon = player:find_target()
   if is_side(target) or is_polygon_floor(target) or
      is_polygon_ceiling(target) or
      is_scenery(target) or is_polygon(target) then
      anchorPlayer(player, x, y, z, polygon)
   elseif is_monster(target) then
      anchorPlayer(player, x, y, z, polygon)
      anchorPlayerToMonster(player, target)
   end
   player._currentActions[justChoseSurface] = iconCountdown
   illustrateConnection(player)
end

function anchorPlayer(player, x, y, z, polygon)
   local anchor = {}
   anchor.x, anchor.y, anchor.z, anchor.polygon =
      x, y, z, polygon
   player._anchor = anchor
end

function anchorPlayerToMonster(player, monster)
   player._anchor.monster = monster
   if not monster._anchors then
      monster._anchors = {}
   end
   table.insert(monster._anchors, player)
end

function updateMonsterAnchor(player)
   if player._anchor and player._anchor.monster then
      local monster = player._anchor.monster
      player._anchor.x, player._anchor.y,
      player._anchor.z, player._anchor.polygon =
	 monster.x, monster.y, monster.z, monster.polygon
   end
end

function unanchorMonster(monster)
   if monster and monster.valid and monster._anchors then
      local anchors = monster._anchors
      for i,player in ipairs(anchors) do
	 resetAnchor(player)
      end
      monster._anchors = {}
   end
end

function motion(player)
   local target, x, y, z, polygon = player:find_target()
   if not target then
      moveMelded(player)
   elseif is_monster(target) then
      if not player._melding then
	 removeMelding(player)
	 meldPlayer(player, target)
      else
	 moveMelded(player)
      end
   else
      changeMeldingCoords(player, x, y, z, polygon)
      moveMelded(player)
   end
   player._currentActions[justMelded] = iconCountdown
end

function moveMelded(player)
   illustrateMelding(player)
   local mx,my,mz =
      advanceConnectionSegment(player.monster.x, player.monster.y,
			       player.monster.z,
			       player._meldingDistance,
			       player.yaw, player.pitch)
   if not mx then return end
   local mPolygon
   for p in Polygons() do
      if p:contains(mx, my, mz) then
	 mPolygon = p
	 break
      end
   end

   -- try to use location of farthest effect
   if not mPolygon then
      mPolygon = player._meldFxPoly
      mx = player._meldFxX
      my = player._meldFxY
      mz = player._meldFxZ - 0.5
   end
   
   if mPolygon then
      changeMeldingCoords(player, mx, my, mz, mPolygon)
   end
end

function findAdjacentPolygon(polygon, line)
   if line.clockwise_polygon == polygon then
      return line.counterclockwise_polygon
   else
      return line.clockwise_polygon
   end
end

function advanceToPolygon(startingPoly, x1, y1, x2, y2)
   local currentPoly = startingPoly
   if currentPoly then
      repeat
	 local line = currentPoly:find_line_crossed_leaving(x1, y1, x2, y2)
	 if line then
	    currentPoly = findAdjacentPolygon(currentPoly, line)
	 end
      until line == nil or currentPoly == nil
   end
   return currentPoly
end

function meldPlayer(player, monster)
   player._melding = monster
   player._meldingDistance = distance3D(
      player.monster.x,
      player.monster.y,
      player.monster.z,
      monster.x,
      monster.y,
      monster.z
   )
   monster._melder = player
end

function removeMelding(player)
   if player._melding then
      player._melding._melder = nil
      player._melding = nil
      player._meldingDistance = nil
   end
end

function changeMeldingCoords(player, x, y, z, polygon)
   local melding = player._melding
   if melding and melding.valid then
      local mPlayer = melding.player
      if not mPlayer then
	 melding:position(x, y, z, polygon)
	 melding.yaw = player.yaw
	 melding.external_velocity = 0
      else
	 local mMelding = mPlayer._melding
	 if
	    mMelding
	    and mMelding.player
	    and mMelding.player == player
	 then
	    removeMelding(player)
	 else
	    mPlayer:position(x, y, z, polygon)
	 end
      end
   end
end

function resetAnchor(player)
   player._anchor = nil
end

function illustrateMelding(player)
   if player._melding and player._melding.valid then
      local x1, y1, z1, poly1, x2, y2, z2, poly2 =
	 player.monster.x, player.monster.y, player.monster.z + 0.5,
      player.monster.polygon,
      player._melding.x, player._melding.y, player._melding.z,
      player._melding.polygon
      local totalDistance = player._meldingDistance
      local effectInterval = totalDistance/connectionSegments/2
      local currentPoly = poly1
      local numberOfEffects = 0
      local yaw, pitch = player.yaw, player.pitch
      local cx, cy, cz = x1, y1, z1
      local nx, ny, nz = advanceConnectionSegment(cx, cy, cz,
						  effectInterval,
						  yaw, pitch)
      while numberOfEffects < connectionSegments do
	 repeat

	    if currentPoly then
	       local line =
		  currentPoly:find_line_crossed_leaving(cx, cy, nx, ny)
	       if not line then
	       elseif line.cw_polygon ~= currentPoly then
		  currentPoly = line.cw_polygon
	       else currentPoly = line.ccw_polygon end
	    end

	 until not currentPoly or currentPoly:contains(nx, ny)
	 if currentPoly then
	    local effect =
	       Effects.new(nx, ny, nz, currentPoly, connectionEffect)
	 end
	 numberOfEffects = numberOfEffects + 1
	 cx, cy, cz = nx, ny, nz
	 nx, ny, nz =
	    advanceConnectionSegment(cx, cy, cz, effectInterval, yaw,
				     pitch)
      end
      player._meldFxPoly = currentPoly
      player._meldFxX = cx
      player._meldFxY = cy
      player._meldFxZ = cz
   end
end

function illustrateConnection(player)
   local x1, y1, z1, poly1, x2, y2, z2, poly2 =
      player.monster.x, player.monster.y, player.monster.z + 0.5,
      player.monster.polygon,
      player._anchor.x, player._anchor.y, player._anchor.z,
      player._anchor.polygon
   local totalDistance = distance3D(x1, y1, z1, x2, y2, z2)
   local effectInterval = totalDistance/connectionSegments
   local currentPoly = poly1
   local numberOfEffects = 0
   local yaw, pitch = findAnglesToAnchor(player)
   local cx, cy, cz = x1, y1, z1
   local nx, ny, nz = advanceConnectionSegment(cx, cy, cz,
					       effectInterval,
					       yaw, pitch)
   while numberOfEffects < connectionSegments do
      repeat

	 if currentPoly then
	    local line = currentPoly:find_line_crossed_leaving(cx, cy, nx, ny)
	    if not line then
	    elseif line.cw_polygon ~= currentPoly then
	       currentPoly = line.cw_polygon
	    else currentPoly = line.ccw_polygon end
	 end

      until not currentPoly or currentPoly:contains(nx, ny)
      if currentPoly then
	 local effect = Effects.new(nx, ny, nz, currentPoly, connectionEffect)
      end
      numberOfEffects = numberOfEffects + 1
      cx, cy, cz = nx, ny, nz
      nx, ny, nz = advanceConnectionSegment(cx, cy, cz, effectInterval, yaw,
					    pitch)
   end
end

function advanceConnectionSegment(x, y, z, distance, angle, elevation)
   angle, elevation = math.rad(angle), math.rad(elevation)
   if distance ~= nil then
   return x + distance * math.cos(elevation) * math.cos(angle),
          y + distance * math.cos(elevation) * math.sin(angle),
          z + distance * math.sin(elevation)
	  
	  end
end

function normalizeAngle(angle)
   while angle < 0 do angle = angle + 360 end
   return angle
end

function showHUD(player)
   local chosen = false
   local status
   for i = justPulled, justReeled, 1 do
      if not player or player.dead then
	 player.overlays[i]:clear()
      else
	 local category = categories[i]
	 chosen = player._currentActions[i]
	 if chosen and chosen > 0 then
	    status = selected
	    player._currentActions[i] = chosen - 1
	 else
	    status = transparent
	    player._currentActions[i] = nil
	 end
	 player.overlays[i].text = " "
	 player.overlays[i].icon =
	    palette..status..icons[i]
	 if not player._anchor and category and justPulled == category then
	    player.overlays[i]:clear()
	 elseif player._anchor and category and justMelded == category then
	    player.overlays[i]:clear()
	 end
      end
      chosen = false
   end
end

palette =
[[
7
.000000
%008000
@33AA33
`BABABA
'D0D0D0
#FFFFFF
]]

transparent =
[[
*FFFFFF00
]]

selected =
[[
*AA2222AA
]]

icons = {}

icons[justMovedCloser] =
[[
****************
****************
****.***********
***.%.**********
**.%%.*****..***
*.%@%....*.'`.**
.%@@%%%%%.#''`.*
.@@@@@@@%###''`.
.@@@@@@@%###''`.
.%@@%%%%%.#''`.*
*.%@%....*.'`.**
**.%%.*****..***
***.%.**********
****.***********
****************
****************
]]

icons[justMovedFarther] =
[[
****************
****************
*****.**********
****.%.*********
****.%%.***..***
*....%@%.*.'`.**
.%%%%%@@%.#''`.*
.%@@@@@@@###''`.
.%@@@@@@@###''`.
.%%%%%@@%.#''`.*
*....%@%.*.'`.**
****.%%.***..***
****.%.*********
*****.**********
****************
****************
]]

icons[justReeled] =
[[
***.%%.*********
**.%@@%.********
*.%@@``%.*******
*.%@``''.*******
.%@``''##...****
.%@`'###''%%.***
.%@`###''%@@%.**
*.%###''%@@@@%.*
*.##'#'`%@@@@%.*
.##%@#'%@@@.@@%.
.#..%%'%@@..@@%.
.##....%@@.@@@%.
*.##.**.%@@@@%.*
**.#.**.%@@@@%.*
*.##.***.%@@%.**
*.#.*****.%%.***
]]

icons[justMelded] =
[[
**...***********
*.@@%.**********
*.@@@%.*********
*.%@@@%.********
**.%@@@%.....***
***.%@@@%@@@@.**
****.%@@@%%%@@.*
*****.%@@%'`%@@.
*****.@%%#''`%@.
*****.@%###''`%.
*****.@%###''`%.
*****.@@%#''`%.*
******.@@%'`%.**
*******.@@%%.***
********....****
****************
]]

icons[justPulled] =
[[
*..............*
.``````````````.
*.''''''''''''`.
**.##########'.*
***.%%@%%@%%#.**
****.%%%@%%..***
*****.%@@%.*****
*****.%@%.******
******.%@%.*****
*****.%@@%.*****
*****.%@%.******
******.%@%.*****
*****.%@@%.*****
*****.%@%..*****
****.%@%@@%.****
*****......*****
]]

icons[justChoseSurface] =
[[
*************..*
************.``.
***********.''`.
**********.##'`.
***********.#'`.
********....#'`.
*.......%@%.#'`.
.%@%@%@%@%@%#'`.
*.......%@%.#'`.
********....#'`.
***********.#'`.
***********.#'`.
***********.#'`.
***********.#..*
************.***
****************
]]
